@@ -1,6 +1,8 @@
 #
 #    Copyright 2017 Mirantis Inc.
 #
+#    Copyright (c) 2022 Dell Inc. or its subsidiaries.
+#
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
 #    a copy of the License at
@@ -13,6 +15,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import ipaddress
 import random
 
 from oslo_utils import uuidutils
@@ -31,7 +34,7 @@ CONF = config.CONF
 class BaremetalStandaloneManager(bm.BaremetalScenarioTest,
                                  manager.NetworkScenarioTest):
 
-    credentials = ['primary', 'admin']
+    credentials = ['primary', 'admin', 'system_admin']
     # NOTE(vsaienko): Standalone tests are using v1/node/<node_ident>/vifs to
     # attach VIF to a node.
     min_microversion = '1.28'
@@ -122,6 +125,16 @@ class BaremetalStandaloneManager(bm.BaremetalScenarioTest,
             node_id, instance_uuid=instance_uuid)
         return body
 
+    @classmethod
+    def _disassociate_instance_with_node(cls, node_id):
+        """Disassociate instance_uuid from attached node.
+
+        :param node_id: Name or UUID of the node.
+        """
+        cls.update_node(node_id, [{'op': 'replace',
+                                   'path': '/instance_uuid',
+                                   'value': None}])
+
     @classmethod
     def get_node_vifs(cls, node_id):
         """Return a list of VIFs for a given node.
@@ -230,9 +243,89 @@ class BaremetalStandaloneManager(bm.BaremetalScenarioTest,
 
         return nodes[0]
 
+    @classmethod
+    def unreserve_node(cls, node):
+        """Unreserves node by disassociating instance_uuid attached to node.
+
+        :param node: Ironic node to disassociate instance_uuid from.
+        """
+
+        def _try_to_disassociate_instance():
+            _, node_prop = cls.baremetal_client.show_node(node['uuid'])
+            if node_prop['instance_uuid']:
+                try:
+                    cls._disassociate_instance_with_node(node['uuid'])
+                except lib_exc.Conflict:
+                    return False
+            return True
+        if (not test_utils.call_until_true(
+                _try_to_disassociate_instance,
+                duration=CONF.baremetal.association_timeout, sleep_for=1)):
+            msg = ('Timed out waiting to disassociate instance from '
+                   'ironic node uuid %s' % node['instance_uuid'])
+            raise lib_exc.TimeoutException(msg)
+
+    @classmethod
+    def gen_config_drive_net_info(cls, node_id, n_port):
+        # Find the port with the vif.
+        use_port = None
+        _, body = cls.baremetal_client.list_node_ports(node_id)
+        for port in body['ports']:
+            _, p = cls.baremetal_client.show_port(port['uuid'])
+            if 'tenant_vif_port_id' in p['internal_info']:
+                use_port = p
+                break
+        if not use_port:
+            m = ('Unable to determine proper mac address to use for config '
+                 'to apply for the virtual media port test.')
+            raise lib_exc.InvalidConfiguration(m)
+        vif_mac_address = use_port['address']
+        if CONF.validation.ip_version_for_ssh == 4:
+            ip_version = "ipv4"
+        else:
+            ip_version = "ipv6"
+        ip_address = n_port['fixed_ips'][0]['ip_address']
+        subnet_id = n_port['fixed_ips'][0]['subnet_id']
+        subnet = cls.os_primary.subnets_client.show_subnet(
+            subnet_id).get('subnet')
+        ip_netmask = str(ipaddress.ip_network(subnet.get('cidr')).netmask)
+        if ip_version == "ipv4":
+            route = [{
+                "netmask": "0.0.0.0",
+                "network": "0.0.0.0",
+                "gateway": subnet.get('gateway_ip'),
+            }]
+        else:
+            # Eh... the data structure doesn't really allow for
+            # this to be easy since a default route with v6
+            # is just referred to as ::/0
+            # so network and netmask would be ::, which is
+            # semi-mind-breaking. Anyway, route advertisers are
+            # expected in this case.
+            route = []
+
+        return {
+            "links": [{"id": "port-test",
+                       "type": "vif",
+                       "ethernet_mac_address": vif_mac_address}],
+            "networks": [
+                {
+                    "id": "network0",
+                    "type": ip_version,
+                    "link": "port-test",
+                    "ip_address": ip_address,
+                    "netmask": ip_netmask,
+                    "network_id": "network0",
+                    "routes": route
+                }
+            ],
+            "services": []
+        }
+
     @classmethod
     def boot_node(cls, image_ref=None, image_checksum=None,
-                  boot_option=None):
+                  boot_option=None, config_drive_networking=False,
+                  fallback_network=None):
         """Boot ironic node.
 
         The following actions are executed:
@@ -248,7 +341,13 @@ class BaremetalStandaloneManager(bm.BaremetalScenarioTest,
         :param boot_option: The defaut boot option to utilize. If not
                             specified, the ironic deployment default shall
                             be utilized.
+        :param config_drive_networking: If we should load configuration drive
+            with network_data values.
+        :param fallback_network: Network to use if we are not able to detect
+            a network for use.
         """
+        config_drive = {}
+
         if image_ref is None:
             image_ref = cls.image_ref
         if image_checksum is None:
@@ -257,8 +356,22 @@ class BaremetalStandaloneManager(bm.BaremetalScenarioTest,
             boot_option = cls.boot_option
 
         network, subnet, router = cls.create_networks()
-        n_port = cls.create_neutron_port(network_id=network['id'])
+        try:
+            n_port = cls.create_neutron_port(network_id=network['id'])
+
+        except TypeError:
+            if fallback_network:
+                n_port = cls.create_neutron_port(
+                    network_id=fallback_network)
+            else:
+                raise
         cls.vif_attach(node_id=cls.node['uuid'], vif_id=n_port['id'])
+        config_drive = None
+        if config_drive_networking:
+            config_drive = {}
+            config_drive['network_data'] = cls.gen_config_drive_net_info(
+                cls.node['uuid'], n_port)
+
         patch = [{'path': '/instance_info/image_source',
                   'op': 'add',
                   'value': image_ref}]
@@ -274,9 +387,15 @@ class BaremetalStandaloneManager(bm.BaremetalScenarioTest,
             patch.append({'path': '/instance_info/capabilities',
                           'op': 'add',
                           'value': {'boot_option': boot_option}})
-        # TODO(vsaienko) add testing for custom configdrive
         cls.update_node(cls.node['uuid'], patch=patch)
-        cls.set_node_provision_state(cls.node['uuid'], 'active')
+
+        if not config_drive:
+            cls.set_node_provision_state(cls.node['uuid'], 'active')
+        else:
+            cls.set_node_provision_state(
+                cls.node['uuid'], 'active',
+                configdrive=config_drive)
+
         cls.wait_power_state(cls.node['uuid'],
                              bm.BaremetalPowerStates.POWER_ON)
         cls.wait_provisioning_state(cls.node['uuid'],
@@ -298,7 +417,12 @@ class BaremetalStandaloneManager(bm.BaremetalScenarioTest,
         cls.detach_all_vifs_from_node(node_id, force_delete=force_delete)
 
         if cls.delete_node or force_delete:
-            cls.set_node_provision_state(node_id, 'deleted')
+            node_state = cls.get_node(node_id)['provision_state']
+            if node_state != bm.BaremetalProvisionStates.AVAILABLE:
+                # Check the state before making the call, to permit tests to
+                # drive node into a clean state before exiting the test, which
+                # is needed for some tests because of complex tests.
+                cls.set_node_provision_state(node_id, 'deleted')
             # NOTE(vsaienko) We expect here fast switching from deleted to
             # available as automated cleaning is disabled so poll status
             # each 1s.
@@ -451,13 +575,25 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
     # set via a different test).
     raid_interface = None
 
+    # The management interface to use by the HW type. The management interface
+    # of the node used in the test will be set to this value. If set to None,
+    # the node will retain its existing management_interface value (which may
+    # have been set via a different test).
+    management_interface = None
+
+    # The power interface to use by the HW type. The power interface of the
+    # node used in the test will be set to this value. If set to None, the
+    # node will retain its existing power_interface value (which may have been
+    # set via a different test).
+    power_interface = None
+
     # Boolean value specify if image is wholedisk or not.
     wholedisk_image = None
 
     # If we need to set provision state 'deleted' for the node  after test
     delete_node = True
 
-    mandatory_attr = ['driver', 'image_ref', 'wholedisk_image']
+    mandatory_attr = ['driver', 'image_ref']
 
     node = None
     node_ip = None
@@ -510,14 +646,39 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
                 "in the list of enabled RAID interfaces %(enabled)s" % {
                     'iface': cls.raid_interface,
                     'enabled': CONF.baremetal.enabled_raid_interfaces})
-        if not cls.wholedisk_image and CONF.baremetal.use_provision_network:
+        if (cls.management_interface and cls.management_interface not in
+                CONF.baremetal.enabled_management_interfaces):
+            raise cls.skipException(
+                "Management interface %(iface)s required by test is not "
+                "in the list of enabled management interfaces %(enabled)s" % {
+                    'iface': cls.management_interface,
+                    'enabled': CONF.baremetal.enabled_management_interfaces})
+        if (cls.power_interface and cls.power_interface not in
+                CONF.baremetal.enabled_power_interfaces):
+            raise cls.skipException(
+                "Power interface %(iface)s required by test is not "
+                "in the list of enabled power interfaces %(enabled)s" % {
+                    'iface': cls.power_interface,
+                    'enabled': CONF.baremetal.enabled_power_interfaces})
+        if (cls.wholedisk_image is not None
+                and not cls.wholedisk_image
+                and CONF.baremetal.use_provision_network):
+            # We only want to enter here if cls.wholedisk_image is set to
+            # a value. If None, skip, if True/False go from there.
             raise cls.skipException(
                 'Partitioned images are not supported with multitenancy.')
 
     @classmethod
-    def set_node_to_active(cls, image_ref=None, image_checksum=None):
-        cls.boot_node(image_ref, image_checksum)
-        if CONF.validation.connect_method == 'floating':
+    def set_node_to_active(cls, image_ref=None, image_checksum=None,
+                           fallback_network=None,
+                           config_drive_networking=None,
+                           method_to_get_ip=None):
+        cls.boot_node(image_ref, image_checksum,
+                      fallback_network=fallback_network,
+                      config_drive_networking=config_drive_networking)
+        if method_to_get_ip:
+            cls.node_ip = method_to_get_ip(cls.node['uuid'])
+        elif CONF.validation.connect_method == 'floating':
             cls.node_ip = cls.add_floatingip_to_node(cls.node['uuid'])
         elif CONF.validation.connect_method == 'fixed':
             cls.node_ip = cls.get_server_ip(cls.node['uuid'])
@@ -548,17 +709,17 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
             boot_kwargs['boot_interface'] = cls.boot_interface
         if cls.raid_interface:
             boot_kwargs['raid_interface'] = cls.raid_interface
+        if cls.management_interface:
+            boot_kwargs['management_interface'] = cls.management_interface
+        if cls.power_interface:
+            boot_kwargs['power_interface'] = cls.power_interface
 
         # just get an available node
         cls.node = cls.get_and_reserve_node()
         cls.update_node_driver(cls.node['uuid'], cls.driver, **boot_kwargs)
 
     @classmethod
-    def resource_cleanup(cls):
-        if CONF.validation.connect_method == 'floating':
-            if cls.node_ip:
-                cls.cleanup_floating_ip(cls.node_ip)
-
+    def cleanup_vif_attachments(cls):
         vifs = cls.get_node_vifs(cls.node['uuid'])
         # Remove ports before deleting node, to catch regression for cases
         # when user did this prior unprovision node.
@@ -567,7 +728,20 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
                 cls.ports_client.delete_port(vif)
             except lib_exc.NotFound:
                 pass
+
+    @classmethod
+    def resource_cleanup(cls):
+        if CONF.validation.connect_method == 'floating':
+            if cls.node_ip:
+                try:
+                    cls.cleanup_floating_ip(cls.node_ip)
+                except IndexError:
+                    # There is no fip to actually remove in this case.
+                    pass
+
+        cls.cleanup_vif_attachments()
         cls.terminate_node(cls.node['uuid'])
+        cls.unreserve_node(cls.node)
         base.reset_baremetal_api_microversion()
         super(BaremetalStandaloneManager, cls).resource_cleanup()
 
@@ -577,7 +751,8 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
         self.assertTrue(self.ping_ip_address(self.node_ip,
                                              should_succeed=should_succeed))
 
-    def build_raid_and_verify_node(self, config=None, deploy_time=False):
+    def build_raid_and_verify_node(self, config=None, deploy_time=False,
+                                   erase_device_metadata=True):
         config = config or self.raid_config
         if deploy_time:
             steps = [
@@ -605,15 +780,15 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
                     "interface": "raid",
                     "step": "delete_configuration"
                 },
-                {
-                    "interface": "deploy",
-                    "step": "erase_devices_metadata",
-                },
                 {
                     "interface": "raid",
                     "step": "create_configuration",
                 }
             ]
+            if erase_device_metadata:
+                steps.insert(1, {
+                    "interface": "deploy",
+                    "step": "erase_devices_metadata"})
             self.baremetal_client.set_node_raid_config(self.node['uuid'],
                                                        config)
             self.manual_cleaning(self.node, clean_steps=steps)
@@ -621,12 +796,14 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
         # The node has been changed, anything at this point, we need to back
         # out the raid configuration.
         if not deploy_time:
-            self.addCleanup(self.remove_raid_configuration, self.node)
+            self.addCleanup(self.remove_raid_configuration, self.node,
+                            erase_device_metadata=erase_device_metadata)
 
         # NOTE(dtantsur): this is not required, but it allows us to check that
         # the RAID device was in fact created and is used for deployment.
         patch = [{'path': '/properties/root_device',
-                  'op': 'add', 'value': {'name': '/dev/md0'}}]
+                  'op': 'add', 'value': {
+                      'name': CONF.baremetal.root_device_name}}]
         if deploy_time:
             patch.append({'path': '/instance_info/traits',
                           'op': 'add', 'value': ['CUSTOM_RAID']})
@@ -642,18 +819,18 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
                   'op': 'remove'}]
         self.update_node(self.node['uuid'], patch=patch)
 
-    def remove_raid_configuration(self, node):
+    def remove_raid_configuration(self, node, erase_device_metadata=True):
         self.baremetal_client.set_node_raid_config(node['uuid'], {})
         steps = [
             {
                 "interface": "raid",
                 "step": "delete_configuration",
-            },
-            {
-                "interface": "deploy",
-                "step": "erase_devices_metadata",
             }
         ]
+        if erase_device_metadata:
+            steps.append({
+                "interface": "deploy",
+                "step": "erase_devices_metadata"})
         self.manual_cleaning(node, clean_steps=steps)
 
     def rescue_unrescue(self):
@@ -728,3 +905,79 @@ class BaremetalStandaloneScenarioTest(BaremetalStandaloneManager):
         self.boot_node_ramdisk(ramdisk_ref, iso)
         self.assertTrue(self.ping_ip_address(self.node_ip,
                                              should_succeed=should_succeed))
+
+    @classmethod
+    def boot_node_anaconda(cls, image_ref, kernel_ref, ramdisk_ref,
+                           stage2_ref=None):
+        """Boot ironic using a ramdisk node.
+
+        The following actions are executed:
+          * Create/Pick networks to boot node in.
+          * Create Neutron port and attach it to node.
+          * Update node image_source.
+          * Deploy node.
+          * Wait until node is deployed.
+
+        :param ramdisk_ref: Reference to user image or ramdisk to boot
+                            the node with.
+        :param iso: Boolean, default False, to indicate if the image ref
+                    us actually an ISO image.
+        """
+        if image_ref is None or kernel_ref is None or ramdisk_ref is None:
+            raise cls.skipException('Skipping anaconda tests as an image ref '
+                                    'was not supplied')
+
+        network, subnet, router = cls.create_networks()
+        n_port = cls.create_neutron_port(network_id=network['id'])
+        cls.vif_attach(node_id=cls.node['uuid'], vif_id=n_port['id'])
+        p_root = '/instance_info/'
+        patch = [{'path': p_root + 'image_source',
+                  'op': 'add',
+                  'value': image_ref},
+                 {'path': p_root + 'kernel',
+                  'op': 'add',
+                  'value': kernel_ref},
+                 {'path': p_root + 'ramdisk',
+                  'op': 'add',
+                  'value': ramdisk_ref}]
+        if stage2_ref:
+            patch.append(
+                {
+                    'path': p_root + 'stage2',
+                    'op': 'add',
+                    'value': stage2_ref,
+                }
+            )
+        cls.update_node(cls.node['uuid'], patch=patch)
+        cls.set_node_provision_state(cls.node['uuid'], 'active')
+        if CONF.validation.connect_method == 'floating':
+            cls.node_ip = cls.add_floatingip_to_node(cls.node['uuid'])
+        elif CONF.validation.connect_method == 'fixed':
+            cls.node_ip = cls.get_server_ip(cls.node['uuid'])
+        else:
+            m = ('Configuration option "[validation]/connect_method" '
+                 'must be set.')
+            raise lib_exc.InvalidConfiguration(m)
+        cls.wait_power_state(cls.node['uuid'],
+                             bm.BaremetalPowerStates.POWER_ON)
+
+        if CONF.baremetal.anaconda_exit_test_at == 'heartbeat':
+            cls.wait_for_agent_heartbeat(
+                cls.node['uuid'],
+                timeout=CONF.baremetal.anaconda_active_timeout)
+        elif CONF.baremetal.anaconda_exit_test_at == 'active':
+            cls.wait_provisioning_state(
+                cls.node['uuid'],
+                bm.BaremetalProvisionStates.ACTIVE,
+                timeout=CONF.baremetal.anaconda_active_timeout,
+                interval=30)
+
+    def boot_and_verify_anaconda_node(self,
+                                      image_ref=None,
+                                      kernel_ref=None,
+                                      ramdisk_ref=None,
+                                      stage2_ref=None,
+                                      should_succeed=True):
+        self.boot_node_anaconda(image_ref, kernel_ref, ramdisk_ref)
+        self.assertTrue(self.ping_ip_address(self.node_ip,
+                                             should_succeed=should_succeed))
